"""
An adapter for fetching JSON data.
"""

# pylint: disable=invalid-name

import json
import logging
import urllib.parse
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests_cache
from jsonpath import JSONPath

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Field
from shillelagh.filters import Filter
from shillelagh.lib import SimpleCostModel, analyze
from shillelagh.typing import Maybe, RequestedOrder, Row

_logger = logging.getLogger(__name__)

SUPPORTED_PROTOCOLS = {"http", "https"}
AVERAGE_NUMBER_OF_ROWS = 100


def get_session() -> requests_cache.CachedSession:
    """
    Return a cached session.
    """
    return requests_cache.CachedSession(
        cache_name="generic_json_cache",
        backend="sqlite",
        expire_after=180,
    )


class GenericJSONAPI(Adapter):

    """
    An adapter for fetching JSON data.
    """

    safe = True

    supports_limit = False
    supports_offset = False

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)
        if parsed.scheme not in SUPPORTED_PROTOCOLS:
            return False
        if fast:
            return Maybe

        session = get_session()
        response = session.head(uri)
        return "application/json" in response.headers["content-type"]

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str]:
        parsed = urllib.parse.urlparse(uri)

        path = urllib.parse.unquote(parsed.fragment) or "$[*]"
        uri = urllib.parse.urlunparse(parsed._replace(fragment=""))

        return uri, path

    def __init__(self, uri: str, path: str = "$[*]"):
        super().__init__()

        self.uri = uri
        self.path = path

        self._session = get_session()

        self._set_columns()

    def _set_columns(self) -> None:
        rows = list(self.get_data({}, []))
        column_names = list(rows[0].keys()) if rows else []

        _, order, types = analyze(iter(rows))

        self.columns = {
            column_name: types[column_name](
                filters=[],
                order=order[column_name],
                exact=False,
            )
            for column_name in column_names
            if column_name != "rowid"
        }

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS)

    def get_data(  # pylint: disable=unused-argument
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        response = self._session.get(self.uri)
        payload = response.json()
        if not response.ok:
            raise ProgrammingError(f'Error: {payload["message"]}')

        parser = JSONPath(self.path)
        for i, row in enumerate(parser.parse(payload)):
            row["rowid"] = i
            _logger.debug(row)
            yield {
                k: json.dumps(v) if isinstance(v, (list, dict)) else v
                for k, v in row.items()
            }



"""Helper functions for Shillelagh."""
import base64
import inspect
import itertools
import math
import operator
import pickle
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Type, TypeVar

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ImpossibleFilterError, ProgrammingError
from shillelagh.fields import Field, Float, Integer, Order, String
from shillelagh.filters import (
    Equal,
    Filter,
    Impossible,
    IsNotNull,
    IsNull,
    Like,
    NotEqual,
    Operator,
    Range,
)
from shillelagh.typing import RequestedOrder, Row

DELETED = range(-1, 0)


class RowIDManager:
    """
    A row ID manager that tracks insert and deletes.

    The ``RowIDManager`` should be used with an append-only table structure.
    It assigns a row ID to each row. When a new row is appended it will
    automatically receive a new ID. And when rows are deleted their ID
    gets changed to -1 to indicate the deletion.

    An example:

        >>> data = ["zero", "one", "two"]
        >>> manager = RowIDManager([range(len(data))])

    To insert data:

        >>> data.append("three")
        >>> manager.insert()
        3
        >>> data.append("four")
        >>> manager.insert(10)  # you can specify a row ID
        10
        >>> for row_id, value in zip(manager, data):
        ...     if row_id != -1:
        ...         print(row_id, value)
        0 zero
        1 one
        2 two
        3 three
        10 four

    To delete data:

        >>> manager.delete(data.index("two"))
        >>> print(data)
        ['zero', 'one', 'two', 'three', 'four']
        >>> for row_id, value in zip(manager, data):
        ...     if row_id != -1:
        ...         print(row_id, value)
        0 zero
        1 one
        3 three
        10 four

    """

    def __init__(self, ranges: List[range]):
        if not ranges:
            raise Exception("Argument ``ranges`` cannot be empty")

        self.ranges = ranges

    def __iter__(self):
        yield from itertools.chain(*self.ranges)

    def get_max_row_id(self) -> int:
        """
        Find the maximum row ID.
        """
        return max((r.stop - 1) for r in self.ranges)

    def check_row_id(self, row_id: int) -> None:
        """
        Check if a provided row ID is not being used.
        """
        for range_ in self.ranges:
            if range_.start <= row_id < range_.stop:
                raise Exception(f"Row ID {row_id} already present")

    def insert(self, row_id: Optional[int] = None) -> int:
        """
        Insert a new row ID.
        """
        if row_id is None:
            max_row_id = self.get_max_row_id()
            row_id = max_row_id + 1
        else:
            self.check_row_id(row_id)

        last = self.ranges[-1]
        if last.stop == row_id:
            self.ranges[-1] = range(last.start, row_id + 1)
        else:
            self.ranges.append(range(row_id, row_id + 1))
        return row_id

    def delete(self, row_id: int) -> None:
        """Mark a given row ID as deleted."""
        for i, range_ in enumerate(self.ranges):
            if range_.start <= row_id < range_.stop:
                if range_.start == range_.stop - 1:
                    self.ranges[i] = DELETED
                elif row_id == range_.start:
                    self.ranges[i] = range(range_.start + 1, range_.stop)
                    self.ranges.insert(i, DELETED)
                elif row_id == range_.stop - 1:
                    self.ranges[i] = range(range_.start, range_.stop - 1)
                    self.ranges.insert(i + 1, DELETED)
                else:
                    self.ranges[i] = range(range_.start, row_id)
                    self.ranges.insert(i + 1, range(row_id + 1, range_.stop))
                    self.ranges.insert(i + 1, DELETED)

                return

        raise Exception(f"Row ID {row_id} not found")


def analyze(
    data: Iterator[Row],
) -> Tuple[int, Dict[str, Order], Dict[str, Type[Field]]]:
    """
    Compute number of rows, order, and types from a stream of rows.
    """
    order: Dict[str, Order] = {}
    types: Dict[str, Type[Field]] = {}

    previous_row: Row = {}
    row: Row = {}
    i = -1
    for i, row in enumerate(data):
        for column_name, value in row.items():
            # determine order
            if i > 0:
                previous = previous_row[column_name]
                order[column_name] = update_order(
                    current_order=order.get(column_name, Order.NONE),
                    previous=previous,
                    current=value,
                    num_rows=i + 1,
                )

            # determine types
            if types.get(column_name) == String:
                continue
            if isinstance(value, (str, list, dict)):
                types[column_name] = String
            elif types.get(column_name) == Float:
                continue
            elif isinstance(value, float):
                types[column_name] = Float
            elif types.get(column_name) == Integer:
                continue
            else:
                types[column_name] = Integer

        previous_row = row

    if row and not order:
        order = {column_name: Order.NONE for column_name in row.keys()}

    num_rows = i + 1

    return num_rows, order, types


def update_order(
    current_order: Order,
    previous: Any,
    current: Any,
    num_rows: int,
) -> Order:
    """
    Update the stored order of a given column.

    This is used to analyze the order of columns, by traversing the
    results and checking if their are sorted in any way.
    """
    if num_rows < 2 or previous is None:
        return Order.NONE

    try:
        if num_rows == 2:
            return Order.ASCENDING if current >= previous else Order.DESCENDING
        if (
            current_order == Order.NONE
            or (current_order == Order.ASCENDING and current < previous)
            or (current_order == Order.DESCENDING and current > previous)
        ):
            return Order.NONE
    except TypeError:
        return Order.NONE

    return current_order


def escape(value: str) -> str:
    """Escape single quotes."""
    return value.replace("'", "''")


def unescape(value: str) -> str:
    """Unescape single quotes."""
    return value.replace("''", "'")


def serialize(value: Any) -> str:
    """
    Serialize adapter arguments.

    This function is used with the SQLite backend, in order to serialize
    the arguments needed to instantiate an adapter via a virtual table.
    """
    return escape(base64.b64encode(pickle.dumps(value)).decode())


def deserialize(value: str) -> Any:
    """
    Deserialize adapter arguments.

    This function is used by the SQLite backend, in order to deserialize
    the virtual table definition and instantiate an adapter.
    """
    return pickle.loads(base64.b64decode(unescape(value).encode()))


def build_sql(  # pylint: disable=too-many-locals, too-many-arguments, too-many-branches
    columns: Dict[str, Field],
    bounds: Dict[str, Filter],
    order: List[Tuple[str, RequestedOrder]],
    table: Optional[str] = None,
    column_map: Optional[Dict[str, str]] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    alias: Optional[str] = None,
) -> str:
    """
    Build a SQL query.

    This is used by adapters which use a simplified SQL dialect to fetch data. For
    GSheets a column map is required, since the SQL references columns by label
    ("A", "B", etc.) instead of name.
    """
    sql = "SELECT *"

    if table:
        sql = f"{sql} FROM {table}"
        if alias:
            sql = f"{sql} AS {alias}"

    conditions = []
    for column_name, filter_ in bounds.items():
        if (
            isinstance(filter_, Range)  # pylint: disable=too-many-boolean-expressions
            and filter_.start is not None
            and filter_.end is not None
            and filter_.start == filter_.end
            and filter_.include_start
            and filter_.include_end
        ):
            filter_ = Equal(filter_.start)

        field = columns[column_name]
        id_ = column_map[column_name] if column_map else column_name
        if alias:
            id_ = f"{alias}.{id_}"
        conditions.extend(get_conditions(id_, field, filter_))
    if conditions:
        sql = f"{sql} WHERE {' AND '.join(conditions)}"

    column_order: List[str] = []
    for column_name, requested_order in order:
        id_ = column_map[column_name] if column_map else column_name
        if alias:
            id_ = f"{alias}.{id_}"
        desc = " DESC" if requested_order == Order.DESCENDING else ""
        column_order.append(f"{id_}{desc}")
    if column_order:
        sql = f"{sql} ORDER BY {', '.join(column_order)}"
    if limit is not None:
        sql = f"{sql} LIMIT {limit}"
    if offset is not None:
        sql = f"{sql} OFFSET {offset}"

    return sql


def get_conditions(id_: str, field: Field, filter_: Filter) -> List[str]:
    """
    Build a SQL condition from a column ID and a filter.
    """
    if isinstance(filter_, Impossible):
        raise ImpossibleFilterError()

    if isinstance(filter_, Equal):
        return [f"{id_} = {field.quote(filter_.value)}"]
    if isinstance(filter_, NotEqual):
        return [f"{id_} != {field.quote(filter_.value)}"]
    if isinstance(filter_, Range):
        conditions = []
        if filter_.start is not None:
            operator_ = ">=" if filter_.include_start else ">"
            conditions.append(f"{id_} {operator_} {field.quote(filter_.start)}")
        if filter_.end is not None:
            operator_ = "<=" if filter_.include_end else "<"
            conditions.append(f"{id_} {operator_} {field.quote(filter_.end)}")
        return conditions
    if isinstance(filter_, Like):
        return [f"{id_} LIKE {field.quote(filter_.value)}"]
    if isinstance(filter_, IsNull):
        return [f"{id_} IS NULL"]
    if isinstance(filter_, IsNotNull):
        return [f"{id_} IS NOT NULL"]

    raise ProgrammingError(f"Invalid filter: {filter_}")


def combine_args_kwargs(
    func: Callable[..., Any], *args: Any, **kwargs: Any
) -> Tuple[Any, ...]:
    """
    Combine args and kwargs into args.

    This is needed because we allow users to pass custom kwargs to adapters,
    but when creating the virtual table we serialize only args.
    """
    signature = inspect.signature(func)
    bound_args = signature.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args.args


def is_null(column: Any, _: Any) -> bool:
    """
    Operator for ``IS NULL``.
    """
    return column is None


def is_not_null(column: Any, _: Any) -> bool:
    """
    Operator for ``IS NOT NULL``.
    """
    return column is not None


def filter_data(
    data: Iterator[Row],
    bounds: Dict[str, Filter],
    order: List[Tuple[str, RequestedOrder]],
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Iterator[Row]:
    """
    Apply filtering and sorting to a stream of rows.

    This is used mostly as an exercise. It's probably much more efficient to
    simply declare fields without any filtering/sorting and let the backend
    (SQLite, eg) handle it.
    """
    for column_name, filter_ in bounds.items():

        def apply_filter(
            data: Iterator[Row],
            operator_: Callable[[Any, Any], bool],
            column_name: str,
            value: Any,
        ) -> Iterator[Row]:
            """
            Apply a given filter to an iterator of rows.

            This method is needed because Python uses lazy bindings in
            generator expressions, so we need to create a new scope in
            order to apply several filters that reuse the same variable
            names.
            """
            return (row for row in data if operator_(row[column_name], value))

        if isinstance(filter_, Impossible):
            return
        if isinstance(filter_, Equal):
            data = apply_filter(data, operator.eq, column_name, filter_.value)
        elif isinstance(filter_, NotEqual):
            data = apply_filter(data, operator.ne, column_name, filter_.value)
        elif isinstance(filter_, Range):
            if filter_.start is not None:
                operator_ = operator.ge if filter_.include_start else operator.gt
                data = apply_filter(data, operator_, column_name, filter_.start)
            if filter_.end is not None:
                operator_ = operator.le if filter_.include_end else operator.lt
                data = apply_filter(data, operator_, column_name, filter_.end)
        elif isinstance(filter_, IsNull):
            data = apply_filter(data, is_null, column_name, None)
        elif isinstance(filter_, IsNotNull):
            data = apply_filter(data, is_not_null, column_name, None)
        else:
            raise ProgrammingError(f"Invalid filter: {filter_}")

    if order:
        # in order to sort we need to consume the iterator and load it into
        # memory :(
        rows = list(data)
        for column_name, requested_order in order:
            reverse = requested_order == Order.DESCENDING
            rows.sort(key=operator.itemgetter(column_name), reverse=reverse)
        data = iter(rows)

    data = apply_limit_and_offset(data, limit, offset)

    yield from data


T = TypeVar("T")


def apply_limit_and_offset(
    rows: Iterator[T],
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Iterator[T]:
    """
    Apply limit/offset to a stream of rows.
    """
    if limit is not None or offset is not None:
        start = offset or 0
        end = None if limit is None else start + limit
        rows = itertools.islice(rows, start, end)
    return rows


def SimpleCostModel(rows: int, fixed_cost: int = 0):  # pylint: disable=invalid-name
    """
    A simple model for estimating query costs.

    The model assumes that each filtering operation is O(n), and each sorting
    operation is O(n log n), in addition to a fixed cost.
    """

    def method(
        obj: Any,  # pylint: disable=unused-argument
        filtered_columns: List[Tuple[str, Operator]],
        order: List[Tuple[str, RequestedOrder]],
    ) -> int:
        return int(
            fixed_cost
            + rows * len(filtered_columns)
            + rows
            * math.log2(rows)  # pylint: disable=c-extension-no-member
            * len(order),
        )

    return method


def find_adapter(
    uri: str,
    adapter_kwargs: Dict[str, Any],
    adapters: List[Type[Adapter]],
) -> Tuple[Type[Adapter], Tuple[Any, ...], Dict[str, Any]]:
    """
    Find an adapter that handles a given URI.

    This is done in 2 passes: first the ``supports`` method is called with ``fast=True``.
    If no adapter returns ``True`` we do a second pass on the plugins that returned
    ``None``, passing ``fast=False`` so they can do network requests to better inspect
    the URI.
    """
    candidates = set()

    for adapter in adapters:
        key = adapter.__name__.lower()
        kwargs = adapter_kwargs.get(key, {})
        supported: Optional[bool] = adapter.supports(uri, fast=True, **kwargs)
        if supported:
            args = adapter.parse_uri(uri)
            return adapter, args, kwargs
        if supported is None:
            candidates.add(adapter)

    for adapter in candidates:
        key = adapter.__name__.lower()
        kwargs = adapter_kwargs.get(key, {})
        if adapter.supports(uri, fast=False, **kwargs):
            args = adapter.parse_uri(uri)
            return adapter, args, kwargs

    raise ProgrammingError(f"Unsupported table: {uri}")



# pylint: disable=abstract-method
"""
A dialect that only connects to GSheets.

This dialect was implemented to replace the ``gsheetsdb`` library.
"""
import logging
import urllib.parse
from operator import itemgetter
from typing import Any, Dict, List, Optional, Tuple, cast

import requests
from google.auth.transport.requests import AuthorizedSession
from sqlalchemy.engine.url import URL
from sqlalchemy.pool.base import _ConnectionFairy
from typing_extensions import TypedDict

from shillelagh.adapters.api.gsheets.lib import get_credentials
from shillelagh.backends.apsw.dialects.base import APSWDialect
from shillelagh.exceptions import ProgrammingError

_logger = logging.getLogger(__name__)


class QueryType(TypedDict, total=False):
    """
    Types for parameters in the SQLAlchemy URI query.
    """

    access_token: str
    service_account_file: str
    subject: str
    app_default_credentials: bool


def extract_query(url: URL) -> QueryType:
    """
    Extract the query from the SQLAlchemy URL.
    """
    if url.query:
        parameters = dict(url.query)
    # there's a bug in how SQLAlchemy <1.4 handles URLs without hosts,
    # putting the query string as the host; handle that case here
    elif url.host and url.host.startswith("?"):
        parameters = dict(urllib.parse.parse_qsl(url.host[1:]))  # pragma: no cover
    else:
        parameters = {}

    if "app_default_credentials" in parameters:
        parameters["app_default_credentials"] = parameters[
            "app_default_credentials"
        ].lower() in {"1", "true"}

    return cast(QueryType, parameters)


class APSWGSheetsDialect(APSWDialect):
    """
    Drop-in replacement for gsheetsdb.

    This dialect loads only the "gsheetsapi" adapter. To use it:

        >>> from sqlalchemy.engine import create_engine
        >>> engine = create_engine("gsheets://")

    """

    # This is supported in ``SQLiteDialect``, and equally supported here. See
    # https://docs.sqlalchemy.org/en/14/core/connections.html#caching-for-third-party-dialects
    # for more context.
    supports_statement_cache = True

    name = "gsheets"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        access_token: Optional[str] = None,
        service_account_file: Optional[str] = None,
        service_account_info: Optional[Dict[str, Any]] = None,
        subject: Optional[str] = None,
        catalog: Optional[Dict[str, str]] = None,
        list_all_sheets: bool = False,
        app_default_credentials: bool = False,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        self.access_token = access_token
        self.service_account_file = service_account_file
        self.service_account_info = service_account_info
        self.subject = subject
        self.catalog = catalog or {}
        self.list_all_sheets = list_all_sheets
        self.app_default_credentials = app_default_credentials

    def create_connect_args(self, url: URL) -> Tuple[Tuple[()], Dict[str, Any]]:
        adapter_kwargs: Dict[str, Any] = {
            "access_token": self.access_token,
            "service_account_file": self.service_account_file,
            "service_account_info": self.service_account_info,
            "subject": self.subject,
            "catalog": self.catalog,
            "app_default_credentials": self.app_default_credentials,
        }
        # parameters can be overridden via the query in the URL
        adapter_kwargs.update(extract_query(url))

        return (), {
            "path": ":memory:",
            "adapters": ["gsheetsapi"],
            "adapter_kwargs": {"gsheetsapi": adapter_kwargs},
            "safe": True,
            "isolation_level": self.isolation_level,
        }

    def do_ping(self, dbapi_connection: _ConnectionFairy) -> bool:
        """
        Return Google Sheets API status.
        """
        response = requests.get(
            "https://www.google.com/appsstatus/dashboard/incidents.json",
        )
        payload = response.json()

        updates = [
            update for update in payload if update["service_name"] == "Google Sheets"
        ]
        if not updates:
            return True

        updates.sort(key=itemgetter("modified"), reverse=True)
        latest_update = updates[0]
        status: str = latest_update["most_recent_update"]["status"]

        if status in {"AVAILABLE", "SERVICE_DISRUPTION"}:
            return True

        # in case we don't understand the status, return True to be conservative
        return status != "SERVICE_OUTAGE"

    def get_table_names(  # pylint: disable=unused-argument
        self, connection: _ConnectionFairy, schema: str = None, **kwargs: Any
    ) -> List[str]:
        """
        Return a list of table names.

        This will query for the authenticated user's spreadsheets, and return
        the URL of each sheet in all the spreadsheets. It's also possible to
        specify a "catalog" of URLs, which are also returned using their short
        names.
        """
        table_names = list(self.catalog.keys())

        query = extract_query(connection.url) if hasattr(connection, "url") else {}
        credentials = get_credentials(
            query.get("access_token", self.access_token),
            query.get("service_account_file", self.service_account_file),
            self.service_account_info,
            query.get("subject", self.subject),
            query.get("app_default_credentials", self.app_default_credentials),
        )
        if not (credentials and self.list_all_sheets):
            return table_names

        session = AuthorizedSession(credentials)

        spreadsheet_ids = get_spreadsheet_ids(session)
        for spreadsheet_id in spreadsheet_ids:
            table_names.extend(get_sheet_urls(spreadsheet_id, session))

        return table_names


def get_spreadsheet_ids(session: AuthorizedSession) -> List[str]:
    """
    Return the ID of all spreadsheets that the user has access to.
    """
    url = (
        "https://www.googleapis.com/drive/v3/files?"
        "q=mimeType='application/vnd.google-apps.spreadsheet'"
    )
    _logger.info("GET %s", url)
    response = session.get(url)
    payload = response.json()
    _logger.debug(payload)
    if "error" in payload:
        raise ProgrammingError(payload["error"]["message"])

    return [file["id"] for file in payload["files"]]


def get_sheet_urls(spreadsheet_id: str, session: AuthorizedSession) -> List[str]:
    """
    Return the URL for all sheets in a given spreadsheet.
    """
    response = session.get(
        "https://sheets.googleapis.com/v4/spreadsheets/"
        f"{spreadsheet_id}?includeGridData=false",
    )
    payload = response.json()
    if "error" in payload:
        _logger.warning(
            "Error loading sheets from file: %s",
            payload["error"]["message"],
        )
        return []

    sheet_urls: List[str] = []
    sheets = payload["sheets"]
    for sheet in sheets:
        sheet_id = sheet["properties"]["sheetId"]
        sheet_urls.append(
            "https://docs.google.com/spreadsheets/d/"
            f"{spreadsheet_id}/edit#gid={sheet_id}",
        )

    return sheet_urls



"""
An adapter for retrieving information on running processes and system utilization (CPU,
memory, disks, network, sensors).

See https://github.com/giampaolo/psutil for more information.
"""
import logging
import time
import urllib.parse
from datetime import datetime, timezone
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import psutil

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import DateTime, Field, Float, Order
from shillelagh.filters import Filter
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

AVERAGE_NUMBER_OF_ROWS = 100


class SystemAPI(Adapter):

    """
    An adapter for retrieving system information.
    """

    safe = False

    supports_limit = True
    supports_offset = True

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)
        return parsed.scheme == "system"

    @staticmethod
    def parse_uri(uri: str) -> Union[Tuple[str], Tuple[str, float]]:
        parsed = urllib.parse.urlparse(uri)
        resource = parsed.netloc
        query_string = urllib.parse.parse_qs(parsed.query)

        if "interval" in query_string:
            return (resource, float(query_string["interval"][0]))
        return (resource,)

    def __init__(self, resource: str, interval: float = 1.0):
        super().__init__()

        self.resource = resource
        self.interval = interval

        self._set_columns()

    def _set_columns(self) -> None:
        self.columns: Dict[str, Field] = {
            "timestamp": DateTime(filters=None, order=Order.ASCENDING, exact=False),
        }

        if self.resource == "cpu":
            num_cpus = psutil.cpu_count()
            for i in range(num_cpus):
                self.columns[f"cpu{i}"] = Float(
                    filters=None,
                    order=Order.NONE,
                    exact=False,
                )
        else:
            raise ProgrammingError(f"Unknown resource: {self.resource}")

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        rowid = 0
        while limit is None or rowid < limit:
            if offset is not None:
                time.sleep(self.interval * offset)

            try:
                values = psutil.cpu_percent(interval=self.interval, percpu=True)
            except KeyboardInterrupt:
                return

            row = {
                "rowid": rowid,
                "timestamp": datetime.now(timezone.utc),
            }
            for i, value in enumerate(values):
                row[f"cpu{i}"] = value / 100.0

            _logger.debug(row)
            yield row
            rowid += 1



"""
Tests for the adapter registry.
"""

import pytest

from shillelagh.adapters.file.csvfile import CSVFile
from shillelagh.adapters.registry import AdapterLoader
from shillelagh.exceptions import InterfaceError

from ..fakes import FakeAdapter


def test_registry(registry: AdapterLoader) -> None:
    """
    Basic tests for the adapter registry.
    """
    registry.clear()
    assert registry.loaders == {}

    registry.add("dummy", FakeAdapter)
    adapter = registry.load("dummy")
    assert adapter == FakeAdapter


def test_load_error(registry: AdapterLoader) -> None:
    """
    Test errors when loading an adapter.
    """
    registry.clear()

    def load_error() -> None:
        raise ImportError("Error!")

    registry.loaders["dummy"].append(load_error)
    with pytest.raises(InterfaceError) as excinfo:
        registry.load("dummy")
    assert str(excinfo.value) == "Unable to load adapter dummy"

    registry.add("dummy", FakeAdapter)
    assert len(registry.loaders["dummy"]) == 2
    adapter = registry.load("dummy")
    assert adapter == FakeAdapter


def test_register(registry: AdapterLoader) -> None:
    """
    Test that we can register new adapters.
    """
    registry.clear()

    registry.register("csvfile", "shillelagh.adapters.file.csvfile", "CSVFile")
    adapter = registry.load("csvfile")
    assert adapter == CSVFile

    registry.register("invalid", "shillelagh.adapters.file.csvfile", "WhatFile")
    with pytest.raises(InterfaceError) as excinfo:
        registry.load("invalid")
    assert str(excinfo.value) == "Unable to load adapter invalid"


def test_load_only_requested_adapters(registry: AdapterLoader) -> None:
    """
    Test that we only try to load requested adapters.
    """
    registry.clear()

    def load_error() -> None:
        raise ImportError("Error!")

    registry.add("valid", FakeAdapter)
    registry.loaders["invalid"].append(load_error)

    assert registry.load_all(["valid"]) == {"valid": FakeAdapter}
    with pytest.raises(InterfaceError) as excinfo:
        registry.load_all()
    assert str(excinfo.value) == "Unable to load adapter invalid"



"""
Tests for shillelagh.fields.
"""
import datetime
from typing import Union

import pytest

from shillelagh.adapters.registry import registry
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import (
    Blob,
    Boolean,
    Date,
    DateTime,
    FastISODateTime,
    Field,
    Float,
    IntBoolean,
    Integer,
    ISODate,
    ISODateTime,
    ISOTime,
    Order,
    String,
    StringBlob,
    StringBoolean,
    StringDate,
    StringDateTime,
    StringDuration,
    StringTime,
    Time,
)
from shillelagh.filters import Equal
from shillelagh.types import BINARY, DATETIME, NUMBER, STRING

from .fakes import FakeAdapter


def test_comparison() -> None:
    """
    Test comparing fields.
    """
    field1 = String(filters=[Equal], order=Order.ASCENDING, exact=True)
    field2 = String(filters=[Equal], order=Order.ASCENDING, exact=True)
    field3 = String(filters=[Equal], order=Order.ASCENDING, exact=False)

    assert field1 == field2
    assert field1 != field3
    assert field1 != 42

    assert String(filters=[Equal], order=Order.ANY, exact=True) != Integer(
        filters=[Equal],
        order=Order.ANY,
        exact=True,
    )


def test_integer() -> None:
    """
    Test ``Integer``.
    """
    assert Integer().parse(1) == 1
    assert Integer().parse(None) is None
    assert Integer().format(1) == 1
    assert Integer().format(None) is None
    assert Integer().quote("1") == "1"  # type: ignore
    assert Integer().quote(None) == "NULL"


def test_float() -> None:
    """
    Test ``Float``.
    """
    assert Float().parse(1.0) == 1.0
    assert Float().parse(None) is None
    assert Float().format(1.0) == 1.0
    assert Float().format(None) is None
    assert Float().quote("1.0") == "1.0"  # type: ignore
    assert Float().quote(None) == "NULL"


def test_string() -> None:
    """
    Test ``String``.
    """
    assert String().parse("1.0") == "1.0"
    assert String().parse(None) is None
    assert String().format("test") == "test"
    assert String().format(None) is None
    assert String().quote("1.0") == "'1.0'"
    assert String().quote("O'Malley's") == "'O''Malley''s'"
    assert String().quote(None) == "NULL"


def test_date() -> None:
    """
    Test ``Date``.
    """
    assert Date().parse(datetime.date(2020, 1, 1)) == datetime.date(2020, 1, 1)
    assert Date().parse(None) is None
    assert Date().format(datetime.date(2020, 1, 1)) == datetime.date(2020, 1, 1)
    assert Date().format(None) is None
    assert Date().quote(datetime.date(2020, 1, 1)) == "'2020-01-01'"
    assert Date().quote(None) == "NULL"


def test_isodate() -> None:
    """
    Test ``ISODate``.
    """
    assert ISODate().parse("2020-01-01") == datetime.date(2020, 1, 1)
    assert ISODate().parse(None) is None
    assert ISODate().parse("invalid") is None
    assert ISODate().format(datetime.date(2020, 1, 1)) == "2020-01-01"
    assert ISODate().format(None) is None
    assert ISODate().quote("2020-01-01") == "'2020-01-01'"
    assert ISODate().quote(None) == "NULL"


def test_string_date() -> None:
    """
    Test ``StringDate``.
    """
    assert StringDate().parse("2020-01-01") == datetime.date(2020, 1, 1)
    assert StringDate().parse("2020-01-01T00:00+00:00") == datetime.date(
        2020,
        1,
        1,
    )
    assert StringDate().parse(None) is None
    assert StringDate().parse("invalid") is None


def test_time() -> None:
    """
    Test ``Time``.
    """
    assert Time().parse(
        datetime.time(12, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.time(12, 0, tzinfo=datetime.timezone.utc)
    assert Time().parse(None) is None
    assert Time().format(
        datetime.time(12, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.time(12, 0, tzinfo=datetime.timezone.utc)
    assert Time().format(None) is None
    assert (
        Time().quote(datetime.time(12, 0, tzinfo=datetime.timezone.utc))
        == "'12:00:00+00:00'"
    )
    assert Time().quote(None) == "NULL"


def test_isotime() -> None:
    """
    Test ``ISOTime``.
    """
    assert ISOTime().parse("12:00+00:00") == datetime.time(
        12,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert ISOTime().parse("12:00") == datetime.time(
        12,
        0,
    )
    assert ISOTime().parse(None) is None
    assert ISOTime().parse("invalid") is None
    assert (
        ISOTime().format(datetime.time(12, 0, tzinfo=datetime.timezone.utc))
        == "12:00:00+00:00"
    )
    assert ISOTime().format(None) is None
    assert ISOTime().quote("12:00:00+00:00") == "'12:00:00+00:00'"
    assert ISOTime().quote(None) == "NULL"


def test_string_time() -> None:
    """
    Test ``StringTime``.
    """
    assert StringTime().parse("12:00+00:00") == datetime.time(
        12,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert StringTime().parse("12:00") == datetime.time(
        12,
        0,
    )
    assert StringTime().parse(None) is None
    assert StringTime().parse("invalid") is None


def test_datetime() -> None:
    """
    Test ``DateTime``.
    """
    assert DateTime().parse(
        datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)
    assert DateTime().parse(None) is None
    assert DateTime().format(
        datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)
    assert DateTime().format(None) is None
    assert (
        DateTime().quote(
            datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
        )
        == "'2020-01-01T12:00:00+00:00'"
    )
    assert DateTime().quote(None) == "NULL"


def test_isodatetime() -> None:
    """
    Test ``ISODateTime``.
    """
    assert ISODateTime().parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert ISODateTime().parse("2020-01-01T12:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
    )
    assert ISODateTime().parse(None) is None
    assert ISODateTime().parse("invalid") is None
    assert (
        ISODateTime().format(
            datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
        )
        == "2020-01-01T12:00:00+00:00"
    )
    assert ISODateTime().format(None) is None
    assert (
        ISODateTime().quote("2020-01-01T12:00:00+00:00")
        == "'2020-01-01T12:00:00+00:00'"
    )
    assert ISODateTime().quote(None) == "NULL"


def test_string_datetime() -> None:
    """
    Test ``StringDateTime``.
    """
    assert StringDateTime().parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert StringDateTime().parse("2020-01-01T12:00Z") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert StringDateTime().parse("2020-01-01T12:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
    )
    assert StringDateTime().parse(None) is None
    assert StringDateTime().parse("invalid") is None


def test_boolean() -> None:
    """
    Test ``Boolean``.
    """
    assert Boolean().parse(True) is True
    assert Boolean().parse(False) is False
    assert Boolean().parse(None) is None
    assert Boolean().format(True) is True
    assert Boolean().format(False) is False
    assert Boolean().format(None) is None
    assert Boolean().quote(True) == "TRUE"
    assert Boolean().quote(False) == "FALSE"
    assert Boolean().quote(None) == "NULL"


def test_int_boolean() -> None:
    """
    Test ``IntBoolean``.
    """
    assert IntBoolean().parse(1) is True
    assert IntBoolean().parse(0) is False
    assert IntBoolean().parse(10) is True
    assert IntBoolean().parse(None) is None
    assert IntBoolean().format(True) == 1
    assert IntBoolean().format(False) == 0
    assert IntBoolean().format(None) is None
    assert IntBoolean().quote(1) == "1"
    assert IntBoolean().quote(0) == "0"
    assert IntBoolean().quote(None) == "NULL"


def test_string_boolean() -> None:
    """
    Test ``StringBoolean``.
    """
    assert StringBoolean().parse("TRUE") is True
    assert StringBoolean().parse("FALSE") is False
    with pytest.raises(ValueError) as expected_err:
        StringBoolean().parse("B")
    assert str(expected_err.value) == "invalid truth value b"
    assert StringBoolean().parse(None) is None
    assert StringBoolean().format(True) == "TRUE"
    assert StringBoolean().format(False) == "FALSE"
    assert StringBoolean().format(None) is None
    assert StringBoolean().quote("TRUE") == "TRUE"
    assert StringBoolean().quote("FALSE") == "FALSE"
    assert StringBoolean().quote(None) == "NULL"


def test_blob() -> None:
    """
    Test ``Blob``.
    """
    assert Blob().parse(b"test") == b"test"
    assert Blob().parse(None) is None
    assert Blob().format(b"test") == b"test"
    assert Blob().format(None) is None
    assert Blob().quote(b"test") == "X'74657374'"
    assert Blob().quote(None) == "NULL"


def test_string_blob() -> None:
    """
    Test ``StringBlob``.
    """
    assert StringBlob().parse("74657374") == b"test"
    assert StringBlob().parse(None) is None
    assert StringBlob().format(b"test") == "74657374"
    assert StringBlob().format(None) is None
    assert StringBlob().quote("74657374") == "X'74657374'"
    assert StringBlob().quote(None) == "NULL"


def test_type_code() -> None:
    """
    Test typecodes for Python DB API.
    """
    assert Integer == NUMBER
    assert Float == NUMBER
    assert String == STRING
    assert Date == DATETIME
    assert Time == DATETIME
    assert DateTime == DATETIME
    assert ISODate == DATETIME
    assert ISOTime == DATETIME
    assert ISODateTime == DATETIME
    assert Blob == BINARY
    assert Boolean == NUMBER

    assert NUMBER != 1


def test_string_duration() -> None:
    """
    Test ``StringDuration``.
    """
    assert StringDuration().parse("12:34:56") == datetime.timedelta(
        hours=12,
        minutes=34,
        seconds=56,
    )
    assert StringDuration().parse("12:34:56.789012") == datetime.timedelta(
        hours=12,
        minutes=34,
        seconds=56,
        microseconds=789012,
    )
    assert StringDuration().parse(None) is None
    assert StringDuration().parse("2 days, 4:00:00") == datetime.timedelta(
        days=2,
        hours=4,
    )
    assert (
        StringDuration().format(datetime.timedelta(hours=12, minutes=34, seconds=56))
        == "12:34:56"
    )
    assert (
        StringDuration().format(
            datetime.timedelta(hours=12, minutes=34, seconds=56, microseconds=789012),
        )
        == "12:34:56.789012"
    )
    assert StringDuration().format(None) is None
    assert (
        StringDuration().format(datetime.timedelta(days=2, hours=4))
        == "2 days, 4:00:00"
    )
    assert (
        StringDuration().quote(
            datetime.timedelta(  # type: ignore
                hours=12,
                minutes=34,
                seconds=56,
                microseconds=789012,
            ),
        )
        == "'12:34:56.789012'"
    )
    assert StringDuration().quote(None) == "NULL"

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = "CREATE TABLE test_table (a DURATION)"
    cursor.execute(sql)

    sql = "INSERT INTO test_table (a) VALUES (?)"
    cursor.execute(sql, (datetime.timedelta(hours=1),))

    sql = "SELECT * FROM test_table"
    cursor.execute(sql)
    assert cursor.fetchall() == [(datetime.timedelta(hours=1),)]


def test_polymorphic_field() -> None:
    """
    Test for a polymorphic field.
    """

    class IntegerOrString(Field[Union[int, str], Union[int, str]]):
        """
        A column that can be an integer or a string.
        """

        type = "TEXT"
        db_api_type = "STRING"

    class CustomFakeAdapter(FakeAdapter):

        """
        A simple adapter with an ``IntegerOrString`` column.
        """

        secret = IntegerOrString()

        def __init__(self):
            super().__init__()

            self.data = [
                {"rowid": 0, "name": "Alice", "age": 20, "pets": 0, "secret": 42},
                {"rowid": 1, "name": "Bob", "age": 23, "pets": 3, "secret": "XXX"},
            ]

    registry.add("dummy", CustomFakeAdapter)

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchall() == [(20, "Alice", 0, 42), (23, "Bob", 3, "XXX")]


def test_fastisodatetime() -> None:
    """
    Test ``FastISODateTime``.
    """
    assert FastISODateTime().parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert FastISODateTime().parse(None) is None
    assert FastISODateTime().parse("2020-01-01T12:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
    )

    with pytest.raises(ProgrammingError) as excinfo:
        FastISODateTime().parse("2020-01-01T12:00Z")
    assert str(excinfo.value) == 'Unable to parse "2020-01-01T12:00Z"'

    with pytest.raises(ProgrammingError) as excinfo:
        FastISODateTime().parse("invalid")
    assert str(excinfo.value) == 'Unable to parse "invalid"'



"""
Test compliance with PEP 249 (https://www.python.org/dev/peps/pep-0249/).
"""

from inspect import isfunction, ismethod

import pytest

from shillelagh.backends.apsw import db as dbapi


def test_module_interface() -> None:
    """
    Test constructors, globals, and exceptions.
    """
    # constructors
    assert isfunction(dbapi.connect)

    # globals
    assert dbapi.apilevel == "2.0"
    assert dbapi.threadsafety == 2
    assert dbapi.paramstyle == "qmark"

    # exceptions
    assert issubclass(dbapi.Warning, Exception)
    assert issubclass(dbapi.Error, Exception)
    assert issubclass(dbapi.InterfaceError, dbapi.Error)
    assert issubclass(dbapi.DatabaseError, dbapi.Error)
    assert issubclass(dbapi.DataError, dbapi.DatabaseError)
    assert issubclass(dbapi.OperationalError, dbapi.DatabaseError)
    assert issubclass(dbapi.IntegrityError, dbapi.DatabaseError)
    assert issubclass(dbapi.InternalError, dbapi.DatabaseError)
    assert issubclass(dbapi.ProgrammingError, dbapi.DatabaseError)
    assert issubclass(dbapi.NotSupportedError, dbapi.DatabaseError)


def test_connection() -> None:
    """
    Test that the connection object implements required methods.
    """
    connection = dbapi.connect(":memory:")

    assert ismethod(connection.close)
    assert ismethod(connection.commit)
    assert ismethod(connection.rollback)  # optional
    assert ismethod(connection.cursor)


def test_cursor() -> None:
    """
    Test that the cursor implements required methods.
    """
    connection = dbapi.connect(":memory:")
    cursor = connection.cursor()

    # attributes
    assert cursor.description is None
    assert cursor.rowcount == -1

    cursor.execute("SELECT 1, 'test'")
    assert cursor.description
    assert len(cursor.description) == 2
    assert all(len(sequence) == 7 for sequence in cursor.description)
    assert cursor.rowcount == 1

    # methods
    assert ismethod(cursor.close)
    cursor.close()
    with pytest.raises(dbapi.Error) as excinfo:
        cursor.execute("SELECT 1")
    assert str(excinfo.value) == "Cursor already closed"

    assert ismethod(cursor.execute)
    assert ismethod(cursor.executemany)
    assert ismethod(cursor.fetchone)
    assert ismethod(cursor.fetchmany)
    assert ismethod(cursor.fetchall)
    assert cursor.arraysize == 1
    cursor.arraysize = 2
    assert cursor.arraysize == 2
    assert ismethod(cursor.setinputsizes)
    assert ismethod(cursor.setoutputsizes)


def test_type_objects_and_constructors() -> None:
    """
    Test type objects and constructors.
    """
    assert isfunction(dbapi.Date)
    assert isfunction(dbapi.Time)
    assert isfunction(dbapi.Timestamp)
    assert isfunction(dbapi.DateFromTicks)
    assert isfunction(dbapi.TimeFromTicks)
    assert isfunction(dbapi.TimestampFromTicks)
    assert isfunction(dbapi.Binary)
    assert dbapi.STRING
    assert dbapi.BINARY
    assert dbapi.NUMBER
    assert dbapi.DATETIME
    assert dbapi.ROWID



"""
Tests for shillelagh.backends.apsw.dialects.base.
"""
from unittest import mock

import pytest
from sqlalchemy import MetaData, Table, create_engine, func, select

from shillelagh.adapters.registry import AdapterLoader
from shillelagh.backends.apsw.dialects.base import APSWDialect
from shillelagh.exceptions import ProgrammingError

from ....fakes import FakeAdapter


def test_create_engine(registry: AdapterLoader) -> None:
    """
    Test ``create_engine``.
    """
    registry.add("dummy", FakeAdapter)

    engine = create_engine("shillelagh://")

    table = Table("dummy://", MetaData(bind=engine), autoload=True)
    query = select(
        [func.sum(table.columns.pets)],  # pylint: disable=no-member
        from_obj=table,
    )
    assert query.scalar() == 3


def test_create_engine_no_adapters(registry: AdapterLoader) -> None:
    """
    Test ``create_engine`` with invalid adapter.
    """
    registry.clear()
    engine = create_engine("shillelagh://")

    with pytest.raises(ProgrammingError) as excinfo:
        Table("dummy://", MetaData(bind=engine), autoload=True)
    assert str(excinfo.value) == "Unsupported table: dummy://"


def test_dialect_ping() -> None:
    """
    Test ``do_ping``.
    """
    mock_dbapi_connection = mock.MagicMock()
    dialect = APSWDialect()
    assert dialect.do_ping(mock_dbapi_connection) is True


def test_has_table(registry: AdapterLoader) -> None:
    """
    Test ``has_table``.
    """
    registry.add("dummy", FakeAdapter)

    engine = create_engine("shillelagh://")
    assert engine.has_table("dummy://a")
    assert engine.has_table("dummy://b")
    assert not engine.has_table("funny://b")



# pylint: disable=c-extension-no-member
"""
Tests for shillelagh.backends.apsw.vt.
"""
import datetime
from typing import Any, Dict, Iterable

import apsw
import pytest
from pytest_mock import MockerFixture

from shillelagh.backends.apsw.vt import (
    VTModule,
    VTTable,
    _add_sqlite_constraint,
    convert_rows_from_sqlite,
    convert_rows_to_sqlite,
    get_all_bounds,
    get_limit_offset,
    type_map,
)
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Field, Float, Integer, Order, String
from shillelagh.filters import Equal, Operator

from ...fakes import FakeAdapter


class FakeAdapterNoFilters(FakeAdapter):

    """
    An adapter where columns have no filters.
    """

    age = Float()
    name = String()
    pets = Integer()


class FakeAdapterOnlyEqual(FakeAdapter):

    """
    An adapter where columns can only be filtered via equality.
    """

    age = Float(filters=[Equal], order=Order.NONE, exact=True)
    name = String(filters=[Equal], order=Order.ASCENDING, exact=True)
    pets = Integer()


class FakeAdapterStaticSort(FakeAdapter):

    """
    An adapter with columns having a static order.
    """

    age = Float(filters=[Equal], order=Order.NONE)
    name = String(filters=[Equal], order=Order.ASCENDING)
    pets = Integer()


class FakeAdapterNoColumns(FakeAdapter):

    """
    An adapter without columns.
    """

    def get_columns(self) -> Dict[str, Field]:
        return {}


def test_vt_module() -> None:
    """
    Test ``VTModule``.
    """
    vt_module = VTModule(FakeAdapter)
    create_table, _ = vt_module.Create(None, "", "", "table")
    assert (
        create_table
        == """CREATE TABLE "table" ("age" REAL, "name" TEXT, "pets" INTEGER)"""
    )


def test_virtual_best_index() -> None:
    """
    Test ``BestIndex``.
    """
    adapter = FakeAdapter()
    adapter.supports_limit = True

    table = VTTable(adapter)
    result = table.BestIndex(
        [
            (1, apsw.SQLITE_INDEX_CONSTRAINT_EQ),  # name =
            (2, apsw.SQLITE_INDEX_CONSTRAINT_GT),  # pets >
            (0, apsw.SQLITE_INDEX_CONSTRAINT_LE),  # age <=
            (-1, 73),  # LIMIT
        ],
        [(1, False)],  # ORDER BY name ASC
    )
    assert result == (
        [(0, True), None, (1, True), (2, True)],
        42,
        (
            f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}], "
            f"[0, {apsw.SQLITE_INDEX_CONSTRAINT_LE}], [-1, 73]], "
            "[[1, false]]]"
        ),
        True,
        666,
    )


def test_virtual_best_index_static_order_not_consumed() -> None:
    """
    Test ``BestIndex`` when the adapter cannot consume the order.
    """
    table = VTTable(FakeAdapterStaticSort())
    result = table.BestIndex(
        [
            (1, apsw.SQLITE_INDEX_CONSTRAINT_EQ),  # name =
            (2, apsw.SQLITE_INDEX_CONSTRAINT_GT),  # pets >
            (0, apsw.SQLITE_INDEX_CONSTRAINT_LE),  # age <=
        ],
        [(1, False)],  # ORDER BY name ASC
    )
    assert result == (
        [(0, False), None, None],
        42,
        f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}]], []]",
        True,
        666,
    )


def test_virtual_best_index_static_order_not_consumed_descending() -> None:
    """
    Test ``BestIndex`` when the adapter cannot consume the order.
    """
    table = VTTable(FakeAdapterStaticSort())
    result = table.BestIndex(
        [
            (1, apsw.SQLITE_INDEX_CONSTRAINT_EQ),  # name =
            (2, apsw.SQLITE_INDEX_CONSTRAINT_GT),  # pets >
            (0, apsw.SQLITE_INDEX_CONSTRAINT_LE),  # age <=
        ],
        [(0, True)],  # ORDER BY age DESC
    )
    assert result == (
        [(0, False), None, None],
        42,
        f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}]], []]",
        False,
        666,
    )


def test_virtual_best_index_operator_not_supported() -> None:
    """
    Test ``BestIndex`` with an unsupported operator.
    """
    table = VTTable(FakeAdapter())
    result = table.BestIndex(
        [(1, apsw.SQLITE_INDEX_CONSTRAINT_MATCH)],  # name LIKE?
        [(1, False)],  # ORDER BY name ASC
    )
    assert result == ([None], 42, "[[], [[1, false]]]", True, 666)


def test_virtual_best_index_order_consumed() -> None:
    """
    Test ``BestIndex`` when the adapter can consume the order.
    """
    table = VTTable(FakeAdapter())
    result = table.BestIndex(
        [
            (1, apsw.SQLITE_INDEX_CONSTRAINT_EQ),  # name =
            (2, apsw.SQLITE_INDEX_CONSTRAINT_GT),  # pets >
            (0, apsw.SQLITE_INDEX_CONSTRAINT_LE),  # age <=
        ],
        [(0, True)],  # ORDER BY age DESC
    )
    assert result == (
        [(0, True), None, (1, True)],
        42,
        (
            f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}], "
            f"[0, {apsw.SQLITE_INDEX_CONSTRAINT_LE}]], [[0, true]]]"
        ),
        True,
        666,
    )


def test_virtual_disconnect() -> None:
    """
    Test ``Disconnect``.
    """
    table = VTTable(FakeAdapter())
    table.Disconnect()  # no-op


def test_update_insert_row() -> None:
    """
    Test ``UpdateInsertRow``.
    """
    adapter = FakeAdapter()
    table = VTTable(adapter)

    new_row_id = table.UpdateInsertRow(None, (6, "Charlie", 1))
    assert new_row_id == 2
    assert list(adapter.get_data({}, [])) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 23, "name": "Bob", "pets": 3, "rowid": 1},
        {"age": 6, "name": "Charlie", "pets": 1, "rowid": 2},
    ]

    new_row_id = table.UpdateInsertRow(4, (40, "Dani", 2))
    assert new_row_id == 4
    assert list(adapter.get_data({}, [])) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 23, "name": "Bob", "pets": 3, "rowid": 1},
        {"age": 6, "name": "Charlie", "pets": 1, "rowid": 2},
        {"age": 40, "name": "Dani", "pets": 2, "rowid": 4},
    ]


def test_update_delete_row() -> None:
    """
    Test ``UpdateDeleteRow``.
    """
    adapter = FakeAdapter()
    table = VTTable(adapter)

    table.UpdateDeleteRow(0)
    assert list(adapter.get_data({}, [])) == [
        {"age": 23, "name": "Bob", "pets": 3, "rowid": 1},
    ]


def test_update_change_row() -> None:
    """
    Test ``UpdateChangeRow``.
    """
    adapter = FakeAdapter()
    table = VTTable(adapter)

    table.UpdateChangeRow(1, 1, (24, "Bob", 4))
    assert list(adapter.get_data({}, [])) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 24, "name": "Bob", "pets": 4, "rowid": 1},
    ]

    table.UpdateChangeRow(1, 2, (24, "Bob", 4))
    assert list(adapter.get_data({}, [])) == [
        {"age": 20, "name": "Alice", "pets": 0, "rowid": 0},
        {"age": 24.0, "name": "Bob", "pets": 4, "rowid": 2},
    ]


def test_cursor() -> None:
    """
    Test the cursor.
    """
    table = VTTable(FakeAdapter())
    cursor = table.Open()
    cursor.Filter(42, "[[], []]", [])
    assert cursor.current_row == (0, 20, "Alice", "0")
    assert cursor.Rowid() == 0
    assert cursor.Column(0) == 20

    cursor.Next()
    assert cursor.current_row == (1, 23, "Bob", "3")

    assert not cursor.Eof()
    cursor.Next()
    assert cursor.Eof()
    cursor.Close()


def test_cursor_with_constraints() -> None:
    """
    Test filtering a cursor.
    """
    table = VTTable(FakeAdapter())
    cursor = table.Open()
    cursor.Filter(42, f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}]], []]", ["Alice"])
    assert cursor.current_row == (0, 20, "Alice", "0")

    assert not cursor.Eof()
    cursor.Next()
    assert cursor.Eof()


def test_cursor_with_constraints_invalid_filter() -> None:
    """
    Test passing an invalid constraint to a cursor.
    """
    table = VTTable(FakeAdapter())
    cursor = table.Open()

    with pytest.raises(Exception) as excinfo:
        cursor.Filter(
            42,
            f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_MATCH}]], []]",
            ["Alice"],
        )

    assert str(excinfo.value) == "Invalid constraint passed: 64"


def test_cursor_with_constraints_no_filters() -> None:
    """
    Test passing a constraint to an adapter that cannot be filtered.
    """
    table = VTTable(FakeAdapterNoFilters())
    cursor = table.Open()
    with pytest.raises(Exception) as excinfo:
        cursor.Filter(42, f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_EQ}]], []]", ["Alice"])

    assert str(excinfo.value) == "No valid filter found"


def test_cursor_with_constraints_only_equal() -> None:
    """
    Test passing a constraint not supported by the adapter.
    """
    table = VTTable(FakeAdapterOnlyEqual())
    cursor = table.Open()
    with pytest.raises(Exception) as excinfo:
        cursor.Filter(42, f"[[[1, {apsw.SQLITE_INDEX_CONSTRAINT_GE}]], []]", ["Alice"])

    assert str(excinfo.value) == "No valid filter found"


def test_adapter_with_no_columns() -> None:
    """
    Test creating a table without columns.
    """
    vt_module = VTModule(FakeAdapterNoColumns)
    with pytest.raises(ProgrammingError) as excinfo:
        vt_module.Create(None, "", "", "table")

    assert str(excinfo.value) == "Virtual table table has no columns"


def test_convert_rows_to_sqlite() -> None:
    """
    Test that rows get converted to types supported by SQLite.
    """
    rows: Iterable[Dict[str, Any]] = [
        {
            "INTEGER": 1,
            "REAL": 1.0,
            "TEXT": "test",
            "TIMESTAMP": datetime.datetime(
                2021,
                1,
                1,
                0,
                0,
                tzinfo=datetime.timezone.utc,
            ),
            "DATE": datetime.date(2021, 1, 1),
            "TIME": datetime.time(0, 0, tzinfo=datetime.timezone.utc),
            "BOOLEAN": False,
            "BLOB": b"test",
        },
        {
            "INTEGER": None,
            "REAL": None,
            "TEXT": None,
            "TIMESTAMP": None,
            "DATE": None,
            "TIME": None,
            "BOOLEAN": None,
            "BLOB": None,
        },
    ]
    columns = {k: v() for k, v in type_map.items()}
    assert list(convert_rows_to_sqlite(columns, iter(rows))) == [
        {
            "INTEGER": "1",
            "REAL": 1.0,
            "TEXT": "test",
            "TIMESTAMP": "2021-01-01T00:00:00+00:00",
            "DATE": "2021-01-01",
            "TIME": "00:00:00+00:00",
            "BOOLEAN": 0,
            "BLOB": b"test",
        },
        {
            "INTEGER": None,
            "REAL": None,
            "TEXT": None,
            "TIMESTAMP": None,
            "DATE": None,
            "TIME": None,
            "BOOLEAN": None,
            "BLOB": None,
        },
    ]


def test_convert_rows_from_sqlite() -> None:
    """
    Test that rows get converted from the types supported by SQLite.
    """
    rows: Iterable[Dict[str, Any]] = [
        {
            "INTEGER": 1,
            "REAL": 1.0,
            "TEXT": "test",
            "TIMESTAMP": "2021-01-01T00:00:00+00:00",
            "DATE": "2021-01-01",
            "TIME": "00:00:00+00:00",
            "BOOLEAN": 0,
            "BLOB": b"test",
        },
        {
            "INTEGER": None,
            "REAL": None,
            "TEXT": None,
            "TIMESTAMP": None,
            "DATE": None,
            "TIME": None,
            "BOOLEAN": None,
            "BLOB": None,
        },
    ]
    columns = {k: v() for k, v in type_map.items()}
    assert list(convert_rows_from_sqlite(columns, iter(rows))) == [
        {
            "INTEGER": 1,
            "REAL": 1.0,
            "TEXT": "test",
            "TIMESTAMP": datetime.datetime(
                2021,
                1,
                1,
                0,
                0,
                tzinfo=datetime.timezone.utc,
            ),
            "DATE": datetime.date(2021, 1, 1),
            "TIME": datetime.time(0, 0, tzinfo=datetime.timezone.utc),
            "BOOLEAN": False,
            "BLOB": b"test",
        },
        {
            "INTEGER": None,
            "REAL": None,
            "TEXT": None,
            "TIMESTAMP": None,
            "DATE": None,
            "TIME": None,
            "BOOLEAN": None,
            "BLOB": None,
        },
    ]


def test_add_sqlite_constraint(mocker: MockerFixture) -> None:
    """
    Test ``_add_sqlite_constraint``.
    """
    operator_map: Dict[int, Operator] = {}
    mocker.patch("shillelagh.backends.apsw.vt.operator_map", new=operator_map)

    _add_sqlite_constraint("INVALID", Operator.LIKE)
    assert operator_map == {}

    _add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_EQ", Operator.EQ)
    assert operator_map == {apsw.SQLITE_INDEX_CONSTRAINT_EQ: Operator.EQ}


def test_get_all_bounds() -> None:
    """
    Test ``get_all_bounds``.
    """
    indexes = [
        (-1, 73),  # LIMIT
        (-1, 74),  # OFFSET
        (0, 2),  # EQ
    ]
    constraintargs = [10, 5, "test"]
    columns: Dict[str, Field] = {"a": String()}

    assert get_all_bounds(indexes, constraintargs, columns) == {
        "a": {(Operator.EQ, "test")},
    }


def test_get_limit_offset() -> None:
    """
    Test ``get_limit_offset``.
    """
    indexes = [
        (-1, 73),  # LIMIT
        (-1, 74),  # OFFSET
        (0, 2),  # EQ
    ]
    constraintargs = [10, 5, "test"]

    limit, offset = get_limit_offset(indexes, constraintargs)
    assert limit == 10
    assert offset == 5

    with pytest.raises(Exception) as excinfo:
        get_limit_offset([(-1, 666)], [10])
    assert str(excinfo.value) == "Invalid constraint passed: 666"

    # ``Operator.EQ``, but column index is -1
    limit, offset = get_limit_offset([(-1, 2)], [10])
    assert limit is None
    assert offset is None



"""
An adapter for GitHub.
"""
import logging
import urllib.parse
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests_cache
from jsonpath import JSONPath

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Boolean, Field, Integer, String, StringDateTime
from shillelagh.filters import Equal, Filter
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

PAGE_SIZE = 100


@dataclass
class Column:
    """
    A class to track columns in GitHub resources.
    """

    # The name of the column in the virtual table, eg, ``userid``.
    name: str

    # The JSON path to the value in the GitHub response. Eg, user ID is
    # returned as ``{"user": {"id": 42}}``, so the JSON path would be
    # ``user.id``.
    json_path: str

    # The Shillelagh field to be used for the column.
    field: Field

    # A default value for when the column is not specified. Eg, for ``pulls``
    # the API defaults to showing only PRs with an open state, so we need to
    # default the column to ``all`` to fetch all PRs when state is not
    # specified in the query.
    default: Optional[Filter] = None


# a mapping from the column name (eg, ``userid``) to the path in the JSON
# response (``{"user": {"id": 42}}`` => ``user.id``) together with the field
TABLES: Dict[str, Dict[str, List[Column]]] = {
    "repos": {
        "pulls": [
            Column("url", "html_url", String()),
            Column("id", "id", Integer()),
            Column("number", "number", Integer(filters=[Equal])),
            Column("state", "state", String(filters=[Equal]), Equal("all")),
            Column("title", "title", String()),
            Column("userid", "user.id", Integer()),
            Column("username", "user.login", String()),
            Column("draft", "draft", Boolean()),
            Column("head", "head.ref", String(filters=[Equal])),  # head.label?
            Column("created_at", "created_at", StringDateTime()),
            Column("updated_at", "updated_at", StringDateTime()),
            Column("closed_at", "closed_at", StringDateTime()),
            Column("merged_at", "merged_at", StringDateTime()),
        ],
    },
}


class GitHubAPI(Adapter):

    """
    An adapter for GitHub.
    """

    safe = True

    supports_limit = True
    supports_offset = True

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)

        if parsed.path.count("/") != 4:
            return False

        # pylint: disable=unused-variable
        _, base, owner, repo, resource = parsed.path.rsplit("/", 4)
        return (
            parsed.netloc == "api.github.com"
            and base in TABLES
            and resource in TABLES[base]
        )

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str, str, str]:
        parsed = urllib.parse.urlparse(uri)
        _, base, owner, repo, resource = parsed.path.rsplit("/", 4)
        return (
            base,
            owner,
            repo,
            resource,
        )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        base: str,
        owner: str,
        repo: str,
        resource: str,
        access_token: Optional[str] = None,
    ):
        super().__init__()

        self.base = base
        self.owner = owner
        self.repo = repo
        self.resource = resource
        self.access_token = access_token

        # use a cache for the API requests
        self._session = requests_cache.CachedSession(
            cache_name="github_cache",
            backend="sqlite",
            expire_after=180,
        )

    def get_columns(self) -> Dict[str, Field]:
        return {
            column.name: column.field for column in TABLES[self.base][self.resource]
        }

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        # apply default values
        for column in TABLES[self.base][self.resource]:
            if column.default is not None and column.name not in bounds:
                bounds[column.name] = column.default

        if "number" in bounds:
            number = bounds.pop("number").value  # type: ignore
            return self._get_single_resource(number, limit, offset)

        return self._get_multiple_resources(bounds, limit, offset)

    def _get_single_resource(
        self,
        number: int,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Iterator[Row]:
        """
        Return a specific resource.
        """
        if offset or (limit is not None and limit < 1):
            return

        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        url = (
            f"https://api.github.com/{self.base}/{self.owner}/"
            f"{self.repo}/{self.resource}/{number}"
        )

        _logger.info("GET %s", url)
        response = self._session.get(url, headers=headers)
        payload = response.json()

        row = {
            column.name: JSONPath(column.json_path).parse(payload)[0]
            for column in TABLES[self.base][self.resource]
        }
        row["rowid"] = 0
        _logger.debug(row)
        yield row

    def _get_multiple_resources(
        self,
        bounds: Dict[str, Filter],
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Iterator[Row]:
        """
        Return multiple resources.
        """
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"

        url = f"https://api.github.com/{self.base}/{self.owner}/{self.repo}/{self.resource}"

        # map filters in ``bounds`` to query params
        params = {name: filter_.value for name, filter_ in bounds.items()}  # type: ignore
        params["per_page"] = PAGE_SIZE

        offset = offset or 0
        page = (offset // PAGE_SIZE) + 1
        offset %= PAGE_SIZE

        rowid = 0
        while limit is None or rowid < limit:
            _logger.info("GET %s (page %d)", url, page)
            params["page"] = page
            response = self._session.get(url, headers=headers, params=params)

            payload = response.json()
            if not payload:
                break

            if not response.ok:
                raise ProgrammingError(payload["message"])

            if offset is not None:
                payload = payload[offset:]
                offset = None

            for resource in payload:
                if limit is not None and rowid == limit:
                    # this never happens because SQLite stops consuming from the generator
                    # as soon as the limit is hit
                    break

                row = {
                    column.name: JSONPath(column.json_path).parse(resource)[0]
                    for column in TABLES[self.base][self.resource]
                }
                row["rowid"] = rowid
                _logger.debug(row)
                yield row
                rowid += 1

            page += 1



"""
Tests for the s3select adapter.
"""
# pylint: disable=unused-argument, redefined-outer-name, use-implicit-booleaness-not-comparison

from typing import cast
from unittest.mock import MagicMock
from urllib.parse import urlparse

import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.api.s3select import (
    CSVSerializationType,
    S3SelectAPI,
    get_input_serialization,
)
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.filters import Impossible


def test_get_input_serialization() -> None:
    """
    Tests for ``get_input_serialization``.
    """
    input_serialization = get_input_serialization(urlparse("s3://bucket/sample.csv"))
    assert input_serialization == {
        "CompressionType": "NONE",
        "CSV": {"FileHeaderInfo": "USE"},
    }

    input_serialization = get_input_serialization(
        urlparse(r"s3://bucket/sample.csv?RecordDelimiter=\r\n"),
    )
    assert input_serialization == {
        "CompressionType": "NONE",
        "CSV": {"FileHeaderInfo": "USE", "RecordDelimiter": "\r\n"},
    }

    input_serialization = get_input_serialization(
        urlparse("s3://bucket/sample?format=csv&CompressionType=GZIP"),
    )
    assert input_serialization == {
        "CompressionType": "GZIP",
        "CSV": {"FileHeaderInfo": "USE"},
    }

    with pytest.raises(ProgrammingError) as excinfo:
        get_input_serialization(urlparse("s3://bucket/sample"))
    assert str(excinfo.value) == (
        "Unable to determine file format. You must declare the format in the "
        "SQLAlchemy URI using ``?format={csv,json,parquet}``."
    )

    input_serialization = get_input_serialization(
        urlparse("s3://bucket/sample?format=csv&FileHeaderInfo=IGNORE&foo=bar"),
    )
    assert input_serialization == {
        "CompressionType": "NONE",
        "CSV": {"FileHeaderInfo": "IGNORE"},
    }

    input_serialization = get_input_serialization(
        urlparse("s3://bucket/sample.json?Type=LINES"),
    )
    assert input_serialization == {
        "CompressionType": "NONE",
        "JSON": {"Type": "LINES"},
    }

    input_serialization = get_input_serialization(
        urlparse("s3://bucket/sample.parquet"),
    )
    assert input_serialization == {
        "CompressionType": "NONE",
        "Parquet": {},
    }

    with pytest.raises(ProgrammingError) as excinfo:
        get_input_serialization(urlparse("s3://bucket/sample.nc"))
    assert str(excinfo.value) == (
        'Invalid format "nc". Valid values: csv, json, parquet'
    )


@pytest.fixture
def boto3(mocker: MockerFixture) -> MagicMock:
    """
    Mock the boto3 client.
    """
    boto3 = mocker.patch("shillelagh.adapters.api.s3select.boto3")
    boto3.client().select_object_content.return_value = {
        "ResponseMetadata": {
            "RequestId": "VFC4GMDAHSX1EQAN",
            "HostId": (
                "jbrLapM/xnEetUcFCXIQNSN+QgBRG1CT6biXwIrr25kobqBHpjZUiFG5f6RY6Ao2IehDmJV"
                "VKtE="
            ),
            "HTTPStatusCode": 200,
            "HTTPHeaders": {
                "x-amz-id-2": (
                    "jbrLapM/xnEetUcFCXIQNSN+QgBRG1CT6biXwIrr25kobqBHpjZUiFG5f6RY6Ao2Ieh"
                    "DmJVVKtE="
                ),
                "x-amz-request-id": "VFC4GMDAHSX1EQAN",
                "date": "Wed, 13 Jul 2022 20:51:38 GMT",
                "transfer-encoding": "chunked",
                "server": "AmazonS3",
            },
            "RetryAttempts": 1,
        },
        "Payload": [
            {
                "Records": {"Payload": b'{"Name":"Sam",'},
            },
            {
                "Records": {
                    "Payload": (
                        b'"PhoneNumber":"(949) 555-1234","City":"Irvine",'
                        b'"Occupation":"Solutions Architect"}\n'
                    ),
                },
            },
            {
                "Stats": {
                    "Details": {
                        "BytesScanned": 624,
                        "BytesProcessed": 624,
                        "BytesReturned": 99,
                    },
                },
            },
            {"End": {}},
        ],
    }
    return cast(MagicMock, boto3)


def test_s3select(boto3: MagicMock) -> None:
    """
    Test the adapter.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = 'SELECT * FROM "s3://bucket/file.csv"'
    data = list(cursor.execute(sql))
    assert data == [("Sam", "(949) 555-1234", "Irvine", "Solutions Architect")]


def test_s3select_with_auth(boto3: MagicMock) -> None:
    """
    Test the adapter.
    """
    connection = connect(
        ":memory:",
        adapter_kwargs={
            "s3selectapi": {
                "aws_access_key_id": "XXX",
                "aws_secret_access_key": "YYY",
            },
        },
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "s3://bucket/file.csv"'
    data = list(cursor.execute(sql))
    assert data == [("Sam", "(949) 555-1234", "Irvine", "Solutions Architect")]
    assert boto3.client.called_with(
        "s3",
        aws_access_key_id="XXX",
        aws_secret_access_key="YYY",
    )


def test_drop_table(boto3: MagicMock) -> None:
    """
    Test that dropping the table deletes the object from S3.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = 'DROP TABLE "s3://bucket/file.csv"'
    cursor.execute(sql)
    boto3.client().delete_object.assert_called_with(Bucket="bucket", Key="file.csv")


def test_impossible_condition(boto3: MagicMock) -> None:
    """
    Test for apsw 3.36 where an impossible condition is passed.
    """
    input_serialization: CSVSerializationType = {"CSV": {}, "CompressionType": "NONE"}
    adapter = S3SelectAPI("bucket", "file.csv", input_serialization)
    assert list(adapter.get_data({"City": Impossible()}, [])) == []



"""
Integration tests for GSheets.

These tests operate on a spreadsheet, fetching and changing data.

Uses a private sheet: https://docs.google.com/spreadsheets/d/
    1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit
"""
import datetime
from typing import Any, Dict

import pytest
from dateutil.tz import tzoffset

from shillelagh.adapters.api.gsheets.types import SyncMode
from shillelagh.backends.apsw.db import connect


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
@pytest.mark.parametrize("sync_mode", [SyncMode.BIDIRECTIONAL, SyncMode.BATCH])
def test_simple_sheet(sync_mode: SyncMode, adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test queries against the simple sheet.

        -----------------
        | country | cnt |
        |---------|-----|
        | BR      |   1 |
        | BR      |   3 |
        | IN      |   5 |
        | ZA      |   6 |
        | CR      |  10 |
        -----------------

    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        "1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit"
        f'?sync_mode={sync_mode.value}#gid=0"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 1.0),
        ("BR", 3.0),
        ("IN", 5.0),
        ("ZA", 6.0),
        ("CR", 10.0),
    ]

    sql = f"INSERT INTO {table} (country, cnt) VALUES (?, ?)"
    cursor.execute(sql, ("US", 14))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 1.0),
        ("BR", 3.0),
        ("IN", 5.0),
        ("ZA", 6.0),
        ("CR", 10.0),
        ("US", 14.0),
    ]

    sql = f"UPDATE {table} SET cnt = cnt + 1"
    cursor.execute(sql)
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 2.0),
        ("BR", 4.0),
        ("IN", 6.0),
        ("ZA", 7.0),
        ("CR", 11.0),
        ("US", 15.0),
    ]

    sql = f"DELETE FROM {table} WHERE country = ?"
    cursor.execute(sql, ("US",))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 2.0),
        ("BR", 4.0),
        ("IN", 6.0),
        ("ZA", 7.0),
        ("CR", 11.0),
    ]

    sql = f"UPDATE {table} SET cnt = cnt - 1"
    cursor.execute(sql)
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 1.0),
        ("BR", 3.0),
        ("IN", 5.0),
        ("ZA", 6.0),
        ("CR", 10.0),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
@pytest.mark.parametrize("sync_mode", [SyncMode.BIDIRECTIONAL, SyncMode.BATCH])
def test_2_header_sheet(sync_mode: SyncMode, adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test a sheet where the column names occupy the first 2 rows.
    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        "1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit"
        f'?sync_mode={sync_mode.value}#gid=1077884006"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 1.0),
        ("BR", 3.0),
        ("IN", 5.0),
        ("ZA", 6.0),
        ("CR", 10.0),
    ]
    assert cursor.description is not None
    assert [(t[0], t[1].type) for t in cursor.description] == [
        ("country string", "TEXT"),
        ("cnt number", "REAL"),
    ]

    sql = f'INSERT INTO {table} ("country string", "cnt number") VALUES (?, ?)'
    cursor.execute(sql, ("US", 14))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 1.0),
        ("BR", 3.0),
        ("IN", 5.0),
        ("ZA", 6.0),
        ("CR", 10.0),
        ("US", 14.0),
    ]

    sql = f"""UPDATE {table} SET "cnt number" = "cnt number" + 1"""
    cursor.execute(sql)
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 2.0),
        ("BR", 4.0),
        ("IN", 6.0),
        ("ZA", 7.0),
        ("CR", 11.0),
        ("US", 15.0),
    ]

    sql = f'DELETE FROM {table} WHERE "country string" = ?'
    cursor.execute(sql, ("US",))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 2.0),
        ("BR", 4.0),
        ("IN", 6.0),
        ("ZA", 7.0),
        ("CR", 11.0),
    ]

    sql = f"""UPDATE {table} SET "cnt number" = "cnt number" - 1"""
    cursor.execute(sql)
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("BR", 1.0),
        ("BR", 3.0),
        ("IN", 5.0),
        ("ZA", 6.0),
        ("CR", 10.0),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
@pytest.mark.parametrize("sync_mode", [SyncMode.BIDIRECTIONAL, SyncMode.BATCH])
def test_types_and_nulls(sync_mode: SyncMode, adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test a sheet with all the supported types, including NULLs.
    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        "1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit"
        f'?sync_mode={sync_mode.value}#gid=1648320094"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (
            datetime.datetime(2018, 9, 1, 0, 0, tzinfo=tzoffset(None, -25200)),
            1.0,
            True,
            datetime.date(2018, 1, 1),
            datetime.time(17, 0),
            "test",
        ),
        (
            datetime.datetime(2018, 9, 2, 0, 0, tzinfo=tzoffset(None, -25200)),
            1.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 3, 0, 0, tzinfo=tzoffset(None, -25200)),
            2.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 4, 0, 0, tzinfo=tzoffset(None, -25200)),
            3.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 5, 0, 0, tzinfo=tzoffset(None, -25200)),
            5.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 6, 0, 0, tzinfo=tzoffset(None, -25200)),
            8.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 7, 0, 0, tzinfo=tzoffset(None, -25200)),
            13.0,
            False,
            None,
            None,
            None,
        ),
        (
            datetime.datetime(2018, 9, 8, 0, 0, tzinfo=tzoffset(None, -25200)),
            None,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 9, 0, 0, tzinfo=tzoffset(None, -25200)),
            34.0,
            None,
            None,
            None,
            "test",
        ),
    ]

    sql = (
        f"INSERT INTO {table} "
        "(datetime, number, boolean, date, timeofday, string) "
        "VALUES (?, ?, ?, ?, ?, ?)"
    )
    cursor.execute(
        sql,
        (
            datetime.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
            10,
            True,
            None,
            None,
            "temporary",
        ),
    )
    cursor.execute(
        sql,
        (
            None,
            None,
            None,
            datetime.date(2021, 7, 10),
            datetime.time(12, 0),
            "temporary",
        ),
    )
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (
            datetime.datetime(2018, 9, 1, 0, 0, tzinfo=tzoffset(None, -25200)),
            1.0,
            True,
            datetime.date(2018, 1, 1),
            datetime.time(17, 0),
            "test",
        ),
        (
            datetime.datetime(2018, 9, 2, 0, 0, tzinfo=tzoffset(None, -25200)),
            1.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 3, 0, 0, tzinfo=tzoffset(None, -25200)),
            2.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 4, 0, 0, tzinfo=tzoffset(None, -25200)),
            3.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 5, 0, 0, tzinfo=tzoffset(None, -25200)),
            5.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 6, 0, 0, tzinfo=tzoffset(None, -25200)),
            8.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 7, 0, 0, tzinfo=tzoffset(None, -25200)),
            13.0,
            False,
            None,
            None,
            None,
        ),
        (
            datetime.datetime(2018, 9, 8, 0, 0, tzinfo=tzoffset(None, -25200)),
            None,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 9, 0, 0, tzinfo=tzoffset(None, -25200)),
            34.0,
            None,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2020, 12, 31, 16, 0, tzinfo=tzoffset(None, -28800)),
            10.0,
            True,
            None,
            None,
            "temporary",
        ),
        (
            None,
            None,
            None,
            datetime.date(2021, 7, 10),
            datetime.time(12, 0),
            "temporary",
        ),
    ]

    sql = f'DELETE FROM {table} WHERE "string" = ?'
    cursor.execute(sql, ("temporary",))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (
            datetime.datetime(2018, 9, 1, 0, 0, tzinfo=tzoffset(None, -25200)),
            1.0,
            True,
            datetime.date(2018, 1, 1),
            datetime.time(17, 0),
            "test",
        ),
        (
            datetime.datetime(2018, 9, 2, 0, 0, tzinfo=tzoffset(None, -25200)),
            1.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 3, 0, 0, tzinfo=tzoffset(None, -25200)),
            2.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 4, 0, 0, tzinfo=tzoffset(None, -25200)),
            3.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 5, 0, 0, tzinfo=tzoffset(None, -25200)),
            5.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 6, 0, 0, tzinfo=tzoffset(None, -25200)),
            8.0,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 7, 0, 0, tzinfo=tzoffset(None, -25200)),
            13.0,
            False,
            None,
            None,
            None,
        ),
        (
            datetime.datetime(2018, 9, 8, 0, 0, tzinfo=tzoffset(None, -25200)),
            None,
            False,
            None,
            None,
            "test",
        ),
        (
            datetime.datetime(2018, 9, 9, 0, 0, tzinfo=tzoffset(None, -25200)),
            34.0,
            None,
            None,
            None,
            "test",
        ),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
@pytest.mark.parametrize("sync_mode", [SyncMode.BIDIRECTIONAL, SyncMode.BATCH])
def test_empty_column(sync_mode: SyncMode, adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test queries against a sheet with an empty column in the middle.

        ---------------------------------------------------------
        | one   |                        | two   | three | four |
        |-------|------------------------|-------|-------|------|
        | test  |                        | test  |   1.5 | 10.1 |
        | test2 | this should not appear | test3 |   0.1 | 10.2 |
        ---------------------------------------------------------

    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        "1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit"
        f'?sync_mode={sync_mode.value}#gid=486670973"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("test", "test", 1.5, 10.1),
        ("test2", "test3", 0.1, 10.2),
    ]
    assert cursor.description is not None
    assert [(t[0], t[1].type) for t in cursor.description] == [
        ("one", "TEXT"),
        ("two", "TEXT"),
        ("three", "REAL"),
        ("four", "REAL"),
    ]

    sql = f"INSERT INTO {table} (two) VALUES (?)"
    cursor.execute(sql, ("test4",))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("test", "test", 1.5, 10.1),
        ("test2", "test3", 0.1, 10.2),
        (None, "test4", None, None),
    ]

    sql = f"DELETE FROM {table} WHERE two = ?"
    cursor.execute(sql, ("test4",))
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        ("test", "test", 1.5, 10.1),
        ("test2", "test3", 0.1, 10.2),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
def test_order_by(adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test that ORDER BY works on multiple columns.
    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        '1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit#gid=624495018"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table} ORDER BY bar, foo, baz"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (1.0, 1.0, 2.0),
        (2.0, 1.0, 5.0),
        (3.0, 1.0, 8.0),
        (1.0, 2.0, 3.0),
        (2.0, 2.0, 6.0),
        (3.0, 2.0, 9.0),
        (1.0, 3.0, 1.0),
        (2.0, 3.0, 4.0),
        (3.0, 3.0, 7.0),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
def test_date_time_formats(adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test that we can parse and modify timestamps with different formats.
    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        '1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit#gid=526664434"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=datetime.timezone.utc),
            datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=datetime.timezone.utc),
        ),
        (
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            None,
            None,
        ),
        (
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            None,
            None,
        ),
        (
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            None,
            None,
        ),
    ]

    sql = f'INSERT INTO {table} ("M/d/yyyy H:mm:ss") VALUES (?)'
    cursor.execute(
        sql,
        (datetime.datetime(2021, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),),
    )
    sql = f'SELECT "M/d/yyyy H:mm:ss" FROM {table} WHERE "M/d/yyyy H:mm:ss" IS NOT NULL'
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=datetime.timezone.utc),),
        (datetime.datetime(2021, 1, 1, 12, 0, tzinfo=datetime.timezone.utc),),
    ]

    sql = f'DELETE FROM {table} WHERE "default" IS NULL'
    cursor.execute(sql)
    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.date(2021, 1, 1),
            datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=datetime.timezone.utc),
            datetime.datetime(2020, 12, 31, 12, 34, 56, tzinfo=datetime.timezone.utc),
        ),
        (
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            datetime.date(2021, 1, 2),
            None,
            None,
        ),
        (
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            datetime.date(2021, 1, 3),
            None,
            None,
        ),
        (
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            datetime.date(2021, 1, 4),
            None,
            None,
        ),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
def test_number_formats(adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test reading data from a table with custom number formats.
    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        '1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit#gid=71119348"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (1234600000.0, 0.1, 1230000000.0, -123.0, 123.0, 5.125),
        (None, None, None, 123.0, None, None),
    ]


@pytest.mark.skip("Credentials no longer valid")
@pytest.mark.slow_integration_test
def test_weird_symbols(adapter_kwargs: Dict[str, Any]) -> None:
    """
    Test reading data from a table where the columns have double quotes in their names.
    """
    table = (
        '"https://docs.google.com/spreadsheets/d/'
        '1_rN3lm0R_bU3NemO0s9pbFkY5LQPcuy1pscv8ZXPtg8/edit#gid=767615647"'
    )

    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = f"SELECT * FROM {table}"
    cursor.execute(sql)
    assert cursor.fetchall() == [(1.0, "a", 45.0), (2.0, "b", 1999.0)]
    assert cursor.description is not None
    assert [column[0] for column in cursor.description] == ['foo"', '"bar', 'a"b']



"""
An adapter to S3 files via S3Select.
"""

import json
import logging
import urllib.parse
from pathlib import Path
from typing import Any, Dict, Iterator, List, Literal, Optional, Tuple, Union, cast

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from typing_extensions import TypedDict

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ImpossibleFilterError, ProgrammingError
from shillelagh.fields import Field, Order
from shillelagh.filters import Equal, Filter, IsNotNull, IsNull, NotEqual, Range
from shillelagh.lib import SimpleCostModel, analyze, build_sql
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

# this is just a wild guess; used to estimate query cost
AVERAGE_NUMBER_OF_ROWS = 1000


class CSVSerializationOptionsType(TypedDict, total=False):
    """
    Options for CSV input.
    """

    AllowQuotedRecordDelimiter: bool
    Comments: str
    FieldDelimiter: str
    FileHeaderInfo: Literal["USE", "IGNORE", "NONE"]
    QuoteCharacter: str
    QuoteEscapedCharacter: str
    RecordDelimiter: str


class CSVSerializationType(TypedDict):
    """
    CSV input serialization.
    """

    CompressionType: str
    CSV: CSVSerializationOptionsType


class JSONSerializationOptionsType(TypedDict, total=False):
    """
    Options for JSON input.
    """

    Type: Literal["DOCUMENT", "LINES"]


class JSONSerializationType(TypedDict):
    """
    CSV input serialization.
    """

    CompressionType: str
    JSON: JSONSerializationOptionsType


class ParquetSerializationOptionsType(TypedDict):
    """
    Options for Parquet input.
    """


class ParquetSerializationType(TypedDict):
    """
    CSV input serialization.
    """

    CompressionType: str
    Parquet: ParquetSerializationOptionsType


InputSerializationType = Union[
    CSVSerializationType,
    JSONSerializationType,
    ParquetSerializationType,
]


def unescape_backslash(value: str) -> str:
    r"""
    Unescape backslashes, converting ``\\n`` into ``\n``.

    For fields like ``RecordDelimiter`` where the user passes a value of "\r\n", the
    parameter should be convert to actual carriage return and new line so that S3
    understands it (ie, ``\r\n`` and not ``\\r\\n``).

    Apparently there's no easy way to do that in Python.
    """
    return value.encode("raw_unicode_escape").decode("unicode_escape")


def get_input_serialization(parsed: urllib.parse.ParseResult) -> InputSerializationType:
    """
    Build the input serialization object.
    """
    options: Dict[str, List[str]] = urllib.parse.parse_qs(parsed.query)

    # the serialization format can be passed explicitly or inferred from the file
    # extension
    if "format" in options:
        format_ = options["format"][-1].lower()
    else:
        key = parsed.path.lstrip("/")
        suffix = Path(key).suffix
        if not suffix:
            raise ProgrammingError(
                "Unable to determine file format. You must declare the format in "
                "the SQLAlchemy URI using ``?format={csv,json,parquet}``.",
            )
        format_ = suffix[1:].lower()

    input_serialization: Dict[str, Any] = {
        "CompressionType": options.get("CompressionType", ["NONE"])[-1],
    }

    if format_ == "csv":
        input_serialization["CSV"] = {
            k: unescape_backslash(v[-1])
            for k, v in options.items()
            if k
            in CSVSerializationOptionsType.__annotations__  # pylint: disable=no-member
        }
        if "FileHeaderInfo" not in input_serialization["CSV"]:
            input_serialization["CSV"]["FileHeaderInfo"] = "USE"

        return cast(CSVSerializationType, input_serialization)

    if format_ == "json":
        input_serialization["JSON"] = {
            k: v[-1]
            for k, v in options.items()
            if k
            in JSONSerializationOptionsType.__annotations__  # pylint: disable=no-member
        }

        return cast(JSONSerializationType, input_serialization)

    if format_ == "parquet":
        input_serialization["Parquet"] = {
            k: v[-1]
            for k, v in options.items()
            if k
            in ParquetSerializationOptionsType.__annotations__  # pylint: disable=no-member
        }

        return cast(ParquetSerializationType, input_serialization)

    raise ProgrammingError(
        f'Invalid format "{format_}". Valid values: csv, json, parquet',
    )


class S3SelectAPI(Adapter):

    """
    An adapter to S3 files via S3Select.

    Used to read data from an S3 bucket::

        s3://bucket-name/path/to/file.extension

    Supported formats for the file are CSV, JSON, and Parquet. They can be either
    inferred from the file extension or declared explicitly::

        s3://bucket-name/sample_data.csv
        s3://bucket-name/sample_data?format=csv

    Different formats have additional configuration options::

        <InputSerialization>
            <CompressionType>string</CompressionType>
            <CSV>
                <AllowQuotedRecordDelimiter>boolean</AllowQuotedRecordDelimiter>
                <Comments>string</Comments>
                <FieldDelimiter>string</FieldDelimiter>
                <FileHeaderInfo>string</FileHeaderInfo>
                <QuoteCharacter>string</QuoteCharacter>
                <QuoteEscapeCharacter>string</QuoteEscapeCharacter>
                <RecordDelimiter>string</RecordDelimiter>
            </CSV>
            <JSON>
                <Type>string</Type>
            </JSON>
            <Parquet>
            </Parquet>
        </InputSerialization>

    For example::

        s3://bucket-name/sample.csv?FileHeaderInfo=Use&CompressionType=NONE

    See https://docs.aws.amazon.com/AmazonS3/latest/API/API_SelectObjectContent.html for
    more info.

    """

    safe = True

    supports_limit = True
    supports_offset = False

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = urllib.parse.urlparse(uri)
        return parsed.scheme == "s3"

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str, InputSerializationType, str]:
        parsed = urllib.parse.urlparse(uri)
        path = urllib.parse.unquote(parsed.fragment) or "$"

        bucket = parsed.netloc
        key = parsed.path.lstrip("/")
        input_serialization = get_input_serialization(parsed)

        return (
            bucket,
            key,
            input_serialization,
            path,
        )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        bucket: str,
        key: str,
        input_serialization: InputSerializationType,
        path: str = "$",
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        s3_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()

        self.bucket = bucket
        self.key = key
        self.input_serialization = input_serialization
        self.table_name = path.replace("$", "S3Object")

        if aws_access_key_id and aws_secret_access_key:
            self.s3_client = boto3.client(
                "s3",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
            )
        else:
            self.s3_client = boto3.client(
                "s3",
                config=Config(signature_version=UNSIGNED),
            )
        self.s3_kwargs = s3_kwargs or {}

        self._set_columns()

    def _set_columns(self) -> None:
        rows = list(self._run_query(f"SELECT * FROM {self.table_name} LIMIT 1"))
        column_names = list(rows[0].keys()) if rows else []
        types = analyze(iter(rows))[2]

        self.columns = {
            column_name: types[column_name](
                filters=[Range, Equal, NotEqual, IsNull, IsNotNull],
                order=Order.NONE,
                exact=True,
            )
            for column_name in column_names
        }

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS)

    def _run_query(self, sql: str) -> Iterator[Dict[str, Any]]:
        """
        Run a query and return rows.
        """
        response = self.s3_client.select_object_content(
            Bucket=self.bucket,
            Key=self.key,
            ExpressionType="SQL",
            Expression=sql,
            InputSerialization=self.input_serialization,
            OutputSerialization={"JSON": {}},
            **self.s3_kwargs,
        )

        # a JSON document might be split across different events
        leftover = ""
        for event in response["Payload"]:
            if "Records" in event:
                records = (
                    leftover + event["Records"]["Payload"].decode("utf-8")
                ).splitlines()
                leftover = ""
                for i, record in enumerate(records):
                    try:
                        row = json.loads(record)
                    except json.decoder.JSONDecodeError:
                        leftover = "\n".join(records[i:])
                        break

                    yield cast(Dict[str, Any], row)

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        try:
            sql = build_sql(
                self.columns,
                bounds,
                order,
                table=self.table_name,
                limit=limit,
                alias="s",
            )
        except ImpossibleFilterError:
            return

        rows = self._run_query(sql)
        for i, row in enumerate(rows):
            row["rowid"] = i
            _logger.debug(row)
            yield {
                k: json.dumps(v) if isinstance(v, (list, dict)) else v
                for k, v in row.items()
            }

    def drop_table(self) -> None:
        self.s3_client.delete_object(Bucket=self.bucket, Key=self.key, **self.s3_kwargs)



"""
A simple example showing the generic JSON.

In this example, the JSON response looks like this:

    {
      "domains": [
        {
          "domain": "facebook-domain-verification8n83p6fzqi04yqdhyqb3vjpv73hzp9.com",
          "create_date": "2022-11-02T14:22:04.363028",
          "update_date": "2022-11-02T14:22:04.363030",
          "country": null,
          "isDead": "False",
          "A": null,
          "NS": null,
          "CNAME": null,
          "MX": null,
          "TXT": null
        },
        {
          "domain": "facebook-password-recover.com",
          "create_date": "2022-11-02T14:22:04.363231",
          "update_date": "2022-11-02T14:22:04.363233",
          "country": null,
          "isDead": "False",
          "A": null,
          "NS": null,
          "CNAME": null,
          "MX": null,
          "TXT": null
        },
        ...
      ],
      "total": 742,
      "time": "6",
      "next_page": null
    }

To get the actual data we need to specify the following JSONPath expression:

    $.domains[*]

"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = """
    SELECT * FROM
    "https://api.domainsdb.info/v1/domains/search?domain=facebook&zone=com#$.domains[*]"
    LIMIT 2
    """
    for row in cursor.execute(SQL):
        print(row)
